@@ -1,6 +1,6 @@
 ###############################################################################
 ## ----------------------------------------------------------------------------
-## Utility functions for Many-Core Engine.
+## Utility functions.
 ##
 ###############################################################################
 
@@ -11,7 +11,7 @@ use warnings;
 
 no warnings qw( threads recursion uninitialized );
 
-our $VERSION = '1.699_003';
+our $VERSION = '1.699_004';
 
 ## no critic (BuiltinFunctions::ProhibitStringyEval)
 
@@ -424,11 +424,11 @@ __END__
 
 =head1 NAME
 
-MCE::Util - Utility functions for Many-Core Engine
+MCE::Util - Utility functions
 
 =head1 VERSION
 
-This document describes MCE::Util version 1.699_003
+This document describes MCE::Util version 1.699_004
 
 =head1 SYNOPSIS
 
